//
// gANTlab module for RWG basis functions
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include <complex>
#include <fstream>
#include <algorithm>

#include <iostream>

#include "module_header.h"

#include "module_data.h"
#include "triquad.h"
#include "green.h"
#include "sparse_matrix.h"
#include "vect.h"
#include "fieldx.h"
#include "gauss.h"

using namespace std;



int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;



int version = 1;
char *name = "LoopTree Module";



extern "C" void* new_data(){ return( (void*) new ModuleData ); }

extern "C" void* dpl_data( void *di_ ){ return( (void*) new ModuleData( (ModuleData*)di_ ) ); }

extern "C" void del_data( void *di_ ){ delete (ModuleData*)di_; }


static double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }


const int matrix_para_no = 8;
const char *matrix_para[ matrix_para_no ] =
{
   "OFF file",
   "mesh unit / m",
   "center point x / m",
   "center point y / m",
   "center point z / m",
   "theta / °",
   "phi / °",
   "psi / °"
};

extern "C" void matrix_get_parameter( const MainData *MD, void *data_,
                                      string **param, string **value, int *no )
{
   ModuleData *data = (ModuleData*)data_;
   char sb[50];

   *no = matrix_para_no;
   *param = new string[*no];
   *value = new string[*no];

   int id = 0;

   (*param)[id] = matrix_para[id];
   (*value)[id] = data->off_file_name;
   id++;

   sprintf(sb,"%g",data->mesh_unit);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cp_x);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cp_y);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cp_z);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->theta * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->phi * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->psi * 180.0 / M_PI );  
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

}

extern "C" void matrix_set_parameter( const MainData *MD, void *data_,
                                      const string *param, const string *value, int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j;

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<matrix_para_no; j++ ) if ( matrix_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->off_file_name = value[i]; break; }
         case 1: { data->mesh_unit = atof( value[i].c_str() ); break; }
         case 2: { data->cp_x = atof( value[i].c_str() ); break; }
         case 3: { data->cp_y = atof( value[i].c_str() ); break; }
         case 4: { data->cp_z = atof( value[i].c_str() ); break; }
         case 5: { data->theta = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 6: { data->phi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 7: { data->psi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         default: throw( "unknown parameter" );
      }
   }

}


extern "C" int matrix_geo_setup( const MainData *MD, void *data_,
                                 string **omsg, int *nomsg )
{
   ModuleData *data = (ModuleData*)data_;

   string msg;

   data->read_off_file( msg );

   delete data->grf;
   data->grf = new GREEN( MD->k0 );

   data->update_Iinc();

   data->setup_exvec();


   double elmax, elmin, el;
   elmax = elmin = HYPOT( data->plist[ data->velist[0].p1 ].x
                             - data->plist[ data->velist[0].p2 ].x,
                          data->plist[ data->velist[0].p1 ].y
                             - data->plist[ data->velist[0].p2 ].y,
                          data->plist[ data->velist[0].p1 ].z
                             - data->plist[ data->velist[0].p2 ].z  );
   for ( vEDGE_LIST::iterator vei = data->velist.begin();
         vei != data->velist.end();
         ++vei )
   {
      el = HYPOT( data->plist[ vei->p1 ].x - data->plist[ vei->p2 ].x,
                  data->plist[ vei->p1 ].y - data->plist[ vei->p2 ].y,
                  data->plist[ vei->p1 ].z - data->plist[ vei->p2 ].z  );
      if ( elmax < el ) elmax = el;
      if ( elmin > el ) elmin = el;
   }


   int no_l = 0, no_t = 0;
   for ( LTbfkt_LIST::iterator it = data->LTbfktlist.begin();
         it != data->LTbfktlist.end();
         ++it )
   {
      if      ( it->count > 1  ) ++no_l;
      else if ( it->count == 1 ) ++no_t;
      else  throw("loop-tree error");
#ifdef BOUNDARY_CHECK
      if ( it->count != it->norm.size() || it->count != it->bfkt_in_LT.size() )
      {
         cerr << "Error in LoopTreeModule matrix_geo_setup(): internal line "
              << __LINE__ << "!" << endl;
         exit(1);
      }
#endif
   }


   *nomsg = 7;
   if ( msg != "" ) { *nomsg += 1; }
   *omsg = new string[*nomsg];

   int si = 0;

   if ( msg != "" )
   {
      (*omsg)[si] = "warning: " + msg;
      ++si;
   }

   const int buffsize = 64;
   char buff[buffsize];
   snprintf( buff, buffsize, "min. edge length: %g m", elmin );
   (*omsg)[si] = buff;
   ++si;
   snprintf( buff, buffsize, "max. edge length: %g m", elmax );                
   (*omsg)[si] = buff;
   ++si;
   snprintf( buff, buffsize, "no. of points   : %u", (unsigned)data->plist.size() );
   (*omsg)[si] = buff;
   ++si;
   snprintf( buff, buffsize, "no. of triangles: %u", (unsigned)data->tlist.size() );
   (*omsg)[si] = buff;
   ++si;
   snprintf( buff, buffsize, "no. of RWG edges: %u", (unsigned)data->velist.size() );
   (*omsg)[si] = buff;
   ++si;
   snprintf( buff, buffsize, "no. of loop bfct: %i", no_l );
   (*omsg)[si] = buff;
   ++si;
   snprintf( buff, buffsize, "no. of tree bfct: %i", no_t );
   (*omsg)[si] = buff;
   ++si;

   return( data->velist.size() );
}


extern "C" void matrix_renumber( const MainData *MD, void *data_, int *perm, int *rperm )
{
   ModuleData *data = (ModuleData*)data_;

   LTbfkt_LIST ltb_t;
   ltb_t.clear();

   for ( unsigned int i=0; i<data->LTbfktlist.size(); ++i )
   {
      ltb_t.push_back( data->LTbfktlist[ rperm[i] ] );
   }

   data->LTbfktlist.swap(ltb_t);
   ltb_t.clear();
}


extern "C" void matrix_bounding_box( const MainData *MD, void *data_,
                                     double *cpos_x, double *cpos_y, double *cpos_z,
                                     double *size_x, double *size_y, double *size_z )
{
   ModuleData *data = (ModuleData*)data_;

   if ( data->plist.size() < 1 )
   {
      *cpos_x = MD->cpos_x;
      *cpos_y = MD->cpos_y;
      *cpos_z = MD->cpos_z;

      *size_x = *size_y = *size_z = 0.0;

      return;
   }

   double minx, maxx, miny, maxy, minz, maxz;
   minx = maxx = data->plist[0].x;
   miny = maxy = data->plist[0].y;
   minz = maxz = data->plist[0].z;

   for ( POINT_LIST::iterator it = data->plist.begin()+1;
         it != data->plist.end();
         it++ )
   {
      minx = ( minx < it->x ? minx : it->x );
      miny = ( miny < it->y ? miny : it->y );
      minz = ( minz < it->z ? minz : it->z );

      maxx = ( maxx > it->x ? maxx : it->x );
      maxy = ( maxy > it->y ? maxy : it->y );
      maxz = ( maxz > it->z ? maxz : it->z );
   }

   *size_x = maxx - minx;
   *size_y = maxy - miny;
   *size_z = maxz - minz;

   *cpos_x = 0.5 * ( maxx + minx );
   *cpos_y = 0.5 * ( maxy + miny );
   *cpos_z = 0.5 * ( maxz + minz );
}


#ifdef HAVE_LIBGL
extern "C" void matrix_add_to_OGLlist( const MainData *MD, void *data_, GLuint list )
{
   ModuleData *data = (ModuleData*)data_;

   data->ogl_triangles( list );
}
#endif


extern "C" void matrix_bfkt_bounding_box( const MainData *MD, void *data_, int nbfkt,
                                          int *sdno,
                                          double **cpos_x, double **cpos_y, double **cpos_z,
                                          double **size_x, double **size_y, double **size_z )
{
   ModuleData *data = (ModuleData*)data_;
   double minx, maxx, miny, maxy, minz, maxz;
   POINT *P;

   *sdno = 0;

   *cpos_x = NULL;
   *cpos_y = NULL;
   *cpos_z = NULL;

   *size_x = NULL;
   *size_y = NULL;
   *size_z = NULL;


//
// bfkt
   if ( nbfkt > -1 && nbfkt < (int)data->LTbfktlist.size() )
   {

   vEDGE *RWG;


   *sdno = data->LTbfktlist[nbfkt].count;

   *cpos_x = new double[*sdno];
   *cpos_y = new double[*sdno];
   *cpos_z = new double[*sdno];

   *size_x = new double[*sdno];
   *size_y = new double[*sdno];
   *size_z = new double[*sdno];


   for ( int s=0; s<*sdno; ++s )
   {
      RWG = &data->velist[ data->LTbfktlist[nbfkt].bfkt_in_LT[s] ];

      P = &data->plist[ RWG->p1 ];
      minx = maxx = P->x;
      miny = maxy = P->y;
      minz = maxz = P->z;

      P = &data->plist[ RWG->p2 ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );

      P = &data->plist[ RWG->spp ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );

      P = &data->plist[ RWG->spm ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );


      (*size_x)[s] = maxx - minx;
      (*size_y)[s] = maxy - miny;
      (*size_z)[s] = maxz - minz;

      (*cpos_x)[s] = 0.5 * ( maxx + minx );
      (*cpos_y)[s] = 0.5 * ( maxy + miny );
      (*cpos_z)[s] = 0.5 * ( maxz + minz );
   }

   }
//
// Iinc
   else if ( nbfkt >= (int)data->LTbfktlist.size() &&
             nbfkt < (int)(data->LTbfktlist.size()+data->Iinclist.size()) )
   {

   Iinc *inc = &data->Iinclist[ nbfkt - data->LTbfktlist.size() ];

   *sdno = 1;

   *cpos_x = new double[*sdno];
   *cpos_y = new double[*sdno];
   *cpos_z = new double[*sdno];

   *size_x = new double[*sdno];
   *size_y = new double[*sdno];
   *size_z = new double[*sdno];

   P = &data->plist[ inc->sp ];
   minx = maxx = P->x;
   miny = maxy = P->y;
   minz = maxz = P->z;

   for ( int i=0; i<inc->count; ++i )
   {
      P = &data->plist[ inc->tp1[i] ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );

      P = &data->plist[ inc->tp2[i] ];
      minx = ( minx < P->x ? minx : P->x );
      miny = ( miny < P->y ? miny : P->y );
      minz = ( minz < P->z ? minz : P->z );

      maxx = ( maxx > P->x ? maxx : P->x );
      maxy = ( maxy > P->y ? maxy : P->y );
      maxz = ( maxz > P->z ? maxz : P->z );
   }

   **size_x = maxx - minx;
   **size_y = maxy - miny;
   **size_z = maxz - minz;

   **cpos_x = 0.5 * ( maxx + minx );
   **cpos_y = 0.5 * ( maxy + miny );
   **cpos_z = 0.5 * ( maxz + minz );

   }
   else throw( "invalid basis function number" );

}


extern "C" void matrix_bfkt_moment( const MainData *MD, void *data_,
                                    int ltno, int sdno,
                                    double cpos_x, double cpos_y, double cpos_z,
                                    double len_x, double len_y, double len_z,
                                    int nx, int ny, int nz,
                                    double *mx, double *my, double *mz, double *md )
{
   ModuleData *data = (ModuleData*)data_;
   triquad irule(7);
   int t, veno;
   double ipx, ipy, ipz, dx, dy, dz, powpow, norm;

   *mx = *my = *mz = *md = 0.0;

//
// bfkt
   if ( ltno > -1 && ltno < (int)data->LTbfktlist.size() )
   {

#ifdef BOUNDARY_CHECK
   if ( sdno < 0 || sdno >= (int)data->LTbfktlist[ltno].count )
   {
      cerr << "Error in LoopTreeModule matrix_bfkt_moment(): internal line "
           << __LINE__ << "!" << endl;
      exit(1); 
   }
#endif

   veno = data->LTbfktlist[ltno].bfkt_in_LT[sdno];
   norm = data->LTbfktlist[ltno].norm[sdno];

// + triangle
   t = data->velist[ veno ].trip;
   for ( int i=0; i<irule.no; i++ )
   {

      ipx = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].x
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].x
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].x;

      ipy = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].y
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].y
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].y;

      ipz = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].z
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].z
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].z;


      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow = pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz )
               * irule.weights[i] * norm;

// RWG functions
      *mx += 0.5 * powpow
             * ( ipx - data->plist[ data->velist[ veno ].spp ].x );
      *my += 0.5 * powpow
             * ( ipy - data->plist[ data->velist[ veno ].spp ].y );
      *mz += 0.5 * powpow
             * ( ipz - data->plist[ data->velist[ veno ].spp ].z );

      *md += powpow;
   }

// - triangle
   t = data->velist[ veno ].trim;
   for ( int i=0; i<irule.no; i++ )
   {

      ipx = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].x
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].x
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].x;

      ipy = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].y
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].y
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].y;

      ipz = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].z
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].z
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].z;


      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow = pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz )
               * irule.weights[i] * norm;

// RWG functions
      *mx -= 0.5 * powpow
             * ( ipx - data->plist[ data->velist[ veno ].spm ].x );
      *my -= 0.5 * powpow
             * ( ipy - data->plist[ data->velist[ veno ].spm ].y );
      *mz -= 0.5 * powpow
             * ( ipz - data->plist[ data->velist[ veno ].spm ].z );

      *md -= powpow;
   }

   if ( data->LTbfktlist[ltno].count > 1 )  // loop
   {
      *md = 0.0;
   }

   }
//  
// Iinc
   else if ( ltno >= (int)data->LTbfktlist.size() &&
             ltno < (int)(data->LTbfktlist.size()+data->Iinclist.size()) )
   {

   Iinc *ici = &data->Iinclist[ ltno - data->LTbfktlist.size() ];

   for ( int t=0; t<ici->count; ++t )
   {

      for ( int i=0; i<irule.no; i++ )
      {

         ipx = irule.lambda1[i] * data->plist[ ici->sp ].x
               + irule.lambda2[i] * data->plist[ ici->tp1[t] ].x
               + irule.lambda3[i] * data->plist[ ici->tp2[t] ].x;

         ipy = irule.lambda1[i] * data->plist[ ici->sp ].y
               + irule.lambda2[i] * data->plist[ ici->tp1[t] ].y
               + irule.lambda3[i] * data->plist[ ici->tp2[t] ].y;

         ipz = irule.lambda1[i] * data->plist[ ici->sp ].z
               + irule.lambda2[i] * data->plist[ ici->tp1[t] ].z
               + irule.lambda3[i] * data->plist[ ici->tp2[t] ].z;


         dx = ( ipx - cpos_x ) / len_x;
         dy = ( ipy - cpos_y ) / len_y;
         dz = ( ipz - cpos_z ) / len_z;

         powpow = pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz )
                  * irule.weights[i] * ici->angle[t];

// inc RWG function 1
         *mx += 0.5 * powpow
                * ( ipx - data->plist[ ici->tp1[t] ].x );
         *my += 0.5 * powpow
                * ( ipy - data->plist[ ici->tp1[t] ].y );
         *mz += 0.5 * powpow
                * ( ipz - data->plist[ ici->tp1[t] ].z );

         *md += powpow;

// inc RWG function 2
         *mx += 0.5 * powpow                                
                * ( ipx - data->plist[ ici->tp2[t] ].x );      
         *my += 0.5 * powpow                                
                * ( ipy - data->plist[ ici->tp2[t] ].y );
         *mz += 0.5 * powpow
                * ( ipz - data->plist[ ici->tp2[t] ].z );

         *md += powpow;

      }

   }

   }
   else throw( "invalid basis function number" );

}



// sort only   
class SORTindex
{
   public:
      int e1, e2;

      SORTindex( int e1_=0, int e2_=0 )
      : e1(e1_), e2(e2_)
      {}

      bool operator> ( const SORTindex& sd2 ) const
      {
         if ( e1 > sd2.e1 )
            return(true);
         else if ( e1 == sd2.e1 )
            if ( e2 > sd2.e2 )
               return(true);
         return(false);
      }

      bool operator< ( const SORTindex& sd2 ) const
      {
         if ( e1 < sd2.e1 )
            return(true);  
         else if ( e1 == sd2.e1 )
            if ( e2 < sd2.e2 )
               return(true);
         return(false);
      }

      bool operator== ( const SORTindex& sd2 ) const
      {
         if ( e1 == sd2.e1 && e2 == sd2.e2 ) return(true);
         return(false);
      }

      bool operator!= ( const SORTindex& sd2 ) const
      {
         if ( e1 != sd2.e1 || e2 != sd2.e2 ) return(true);
         return(false);
      }
};

extern "C" void matrix_ele_setup( const MainData *MD, void *data_,
                                  string **omsg, int *nomsg )
{
   ModuleData *data = (ModuleData*)data_;

   unsigned int i, j;
   double r1, r2, D, CP1[3], CP2[3];

   const complex<double> cNULL(0.0,0.0);
   const complex<double> cONE(1.0,0.0);
   const complex<double> cJ(0.0,1.0);

// symbolic
   vector<SORTindex> SI;
   SI.clear();

   for ( i=0; i<data->LTbfktlist.size(); ++i )
   {
      data->umkugel( i, r1, CP1 );

      D = MD->wavelength*MD->near_limit + 1.01*r1;

      for ( j=i; j<data->LTbfktlist.size(); ++j )
      {
         data->umkugel( j, r2, CP2 );

         if ( HYPOT( CP2[0]-CP1[0], CP2[1]-CP1[1], CP2[2]-CP1[2] )
              < D + 1.01*r2 )
         {

            for ( vector<int>::iterator iti=data->LTbfktlist[i].bfkt_in_LT.begin();
                  iti != data->LTbfktlist[i].bfkt_in_LT.end();
                  ++iti )
            {
            for ( vector<int>::iterator itj=data->LTbfktlist[j].bfkt_in_LT.begin();
                  itj != data->LTbfktlist[j].bfkt_in_LT.end();
                  ++itj )
            {
               SI.push_back( SORTindex(*iti,*itj) );
               SI.push_back( SORTindex(*itj,*iti) );
            }
            }

         }
      }

   }

   sort( SI.begin(), SI.end() );
   vector<SORTindex>::iterator sir = unique( SI.begin(), SI.end() );

   data->Zrwg = new SparseMatrix< SEP_Z_VAL >( false,
                                               data->LTbfktlist.size(),
                                               sir-SI.begin() );

   int k, e1;

   k = e1 = 0;
   data->Zrwg->first_ele[0] = 0;

   for ( vector<SORTindex>::iterator its = SI.begin();
         its != sir;
         ++its )
   {
      if ( its->e1 != e1 )
      {
         for ( ++e1; e1 < its->e1; ++e1 )
         {
#ifdef BOUNDARY_CHECK
            if ( e1 < 0 || e1 > data->Zrwg->nofrows )
            {
               cerr << "Error in LoopTreeModule matrix_ele_setup(): internal line "
                    << __LINE__ << "!" << endl;
               exit(1);
            }
#endif
            data->Zrwg->first_ele[e1] = k;
         }

         data->Zrwg->first_ele[e1] = k;
      }

#ifdef BOUNDARY_CHECK  
      if ( k >= data->Zrwg->totlen )
      {
         cerr << "Error in LoopTreeModule matrix_ele_setup(): internal line "
              << __LINE__ << "!" << endl;
         exit(1);
      }
#endif    
      data->Zrwg->colid[k] = its->e2;
      ++k;
   }

   SI.clear();

// numeric
   triquad oI(3), iI(7);

   bool singu;
   int atri, qtri, e2;
   double rv1[3], rv2[3], area;
   complex<double> A[3], phi, Gd[3][3], Gs, v[3], J[3], div, J0[3], div0;
   POINT ap, qp;

   complex<double> jomega = cJ*MD->omega;

#pragma omp parallel for private(k,singu,atri,qtri,e2,rv1,rv2,area,A,phi,Gd,Gs,v,J,div,J0,div0,ap,qp)
   for ( e1=0; e1<data->Zrwg->nofrows; ++e1 )
   {
   for ( k=data->Zrwg->first_ele[e1]; k<data->Zrwg->first_ele[e1+1]; ++k )
   {
   e2 = data->Zrwg->colid[k];

   data->Zrwg->ele[k] = cNULL;

// triangle 1 p
   atri = data->velist[e1].trip;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].y
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].y
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].z
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].z
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].z;


      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi = cNULL;

// triangle 2 p
      qtri = data->velist[e2].trip;

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( ap.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( ap.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
         rv2[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
         rv2[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;

         div = cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

// triangle 2 m
      qtri = data->velist[e2].trim;

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( data->plist[ data->velist[e2].spm ].x - ap.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - ap.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - ap.z ) * 0.5;
            div0 = -cONE;
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;
            div0 = -cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;

         rv2[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
         rv2[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
         rv2[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;

         div = -cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

      rv1[0] = ( ap.x - data->plist[ data->velist[e1].spp ].x ) * 0.5;
      rv1[1] = ( ap.y - data->plist[ data->velist[e1].spp ].y ) * 0.5;
      rv1[2] = ( ap.z - data->plist[ data->velist[e1].spp ].z ) * 0.5;

      data->Zrwg->ele[k] += SEP_Z_VAL(
                       - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] ),
                       - 1.0 * phi / jomega
                                     ) * oI.weights[ip1];
   }

// triangle 1 m
   atri = data->velist[e1].trim;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].y
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].y
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].z
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].z
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].z;


      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi = cNULL;

// triangle 2 p
      qtri = data->velist[e2].trip;

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( ap.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( ap.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
         rv2[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
         rv2[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;

         div = cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

// triangle 2 m
      qtri = data->velist[e2].trim;

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( data->plist[ data->velist[e2].spm ].x - ap.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - ap.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - ap.z ) * 0.5;

            div0 = -cONE;
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;

            div0 = -cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
         rv2[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
         rv2[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;

         div = -cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

      rv1[0] = ( data->plist[ data->velist[e1].spm ].x - ap.x ) * 0.5;
      rv1[1] = ( data->plist[ data->velist[e1].spm ].y - ap.y ) * 0.5;
      rv1[2] = ( data->plist[ data->velist[e1].spm ].z - ap.z ) * 0.5;

      data->Zrwg->ele[k] += SEP_Z_VAL(
                       - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] ),
                       - (-1.0) * phi / jomega
                                     ) * oI.weights[ip1];
   }

   }  // end:e2
   }  // end:e1

   *nomsg = 0;
   *omsg = NULL;
}


extern "C" complex<double> matrix_ele( const MainData *MD, void *data_, int e1, int e2 )
{
   ModuleData *data = (ModuleData*)data_;

   const SEP_Z_VAL cNULL(0.0,0.0);
   SEP_Z_VAL cum1, cum2;

   cum1 = cNULL;

   for ( unsigned int i=0; i<data->LTbfktlist[e1].count; ++i )
   {

      cum2 = cNULL;
      for ( unsigned int j=0; j<data->LTbfktlist[e2].count; ++j )
      {
         cum2 += (*data->Zrwg)( data->LTbfktlist[e1].bfkt_in_LT[i],
                                data->LTbfktlist[e2].bfkt_in_LT[j]  )
                 * data->LTbfktlist[e2].norm[j];
      }

      cum1 += cum2 * data->LTbfktlist[e1].norm[i];
   }

   if ( data->LTbfktlist[e1].count > 1 || data->LTbfktlist[e2].count > 1 )
      return( cum1.a_val );  // loop

   return( cum1.a_val + cum1.phi_val );
}


extern "C" void matrix_ele_cleanup( const MainData *MD, void *data_ )
{
   ModuleData *data = (ModuleData*)data_;
   delete data->Zrwg;
   data->Zrwg = NULL;
}





const int RHS_para_no = 1;
const char *RHS_para[ RHS_para_no ] =
{
   "no. of I sources"
};

const int Iinc_para_no = 5;
const char *Iinc_para[ Iinc_para_no ] =
{
   "pos x / m",
   "pos y / m",
   "pos z / m",
   "abs(I) / A",
   "arg(I) / °" 
};

extern "C" void RHS_get_parameter( const MainData *MD, void *data_,
                                   string **param, string **value, int *no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   char sb[buffsize];

   *no = RHS_para_no + data->Iinclist.size()*Iinc_para_no;
   *param = new string[*no];
   *value = new string[*no];
   int id = 0;
   int id2 = 0;

   sprintf(sb,"%u",(unsigned)data->Iinclist.size());
   (*param)[id] = RHS_para[id2];  
   (*value)[id] = sb;
   id++;

   for ( unsigned int w=0; w<data->Iinclist.size(); w++ )
   {
      id2 = 0;
      snprintf( sb, buffsize, "%i:%s", w+1, Iinc_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->Iinclist[w].x );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, Iinc_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->Iinclist[w].y );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", w+1, Iinc_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->Iinclist[w].z );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", w+1, Iinc_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", abs(data->Iinclist[w].Iampl) );
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", w+1, Iinc_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", arg(data->Iinclist[w].Iampl)*180.0/M_PI );
      (*value)[id] = sb;
      id2++;
      id++;
   }

}

extern "C" void RHS_set_parameter( const MainData *MD, void *data_,
                                   const string *param, const string *value, int no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 100;
   char sb[buffsize];
   int j;
   unsigned int ni;
   bool sec;

   ni = data->Iinclist.size();

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<RHS_para_no; j++ ) if ( RHS_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { ni = atoi( value[i].c_str() ); break; }
      }
   }

// Iinc
   if ( ni != data->Iinclist.size() ) data->Iinclist.resize(ni);

   for ( unsigned int g=0; g<data->Iinclist.size(); ++g )
   {
      double absv = abs(data->Iinclist[g].Iampl);
      double argv = arg(data->Iinclist[g].Iampl);

      sec = true;
      for ( int i=0; i<no; i++ )
      {
         if ( RHS_para[0] == param[i] )   // Iinc
         {
            sec = false;
            continue;   
         }
         if ( sec ) continue;

         for ( j=0; j<Iinc_para_no; j++ )
         {
            snprintf( sb, buffsize, "%i:%s", g+1, Iinc_para[j] );

            if ( sb == param[i] )
            {
               switch(j)
               {
                  case 0: { data->Iinclist[g].x = atof( value[i].c_str() ); break; }
                  case 1: { data->Iinclist[g].y = atof( value[i].c_str() ); break; }
                  case 2: { data->Iinclist[g].z = atof( value[i].c_str() ); break; }
                  case 3: { absv = atof( value[i].c_str() ); break; }
                  case 4: { argv = atof( value[i].c_str() )*M_PI/180.0; break; }
               }
            }   

         }

      }

      data->Iinclist[g].Iampl = absv * complex<double>( cos(argv), sin(argv) );
   }

   data->update_Iinc();
}

#ifdef HAVE_LIBGL
extern "C" void RHS_add_to_OGLlist( const MainData *MD, void *data_, GLuint list )
{
   ModuleData *data = (ModuleData*)data_;
   const GLfloat clr[4]   = { 1.00, 0.00, 1.00, 1.0 };

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );

   glBegin(GL_LINES);

   for ( Iinc_LIST::iterator ici = data->Iinclist.begin();
         ici != data->Iinclist.end();
         ++ici )
   {

      for ( int t=0; t<ici->count; ++t )
      {
         glVertex3d( data->plist[ ici->sp ].x,
                     data->plist[ ici->sp ].y,
                     data->plist[ ici->sp ].z  );
         glVertex3d( data->plist[ ici->tp1[t] ].x,
                     data->plist[ ici->tp1[t] ].y,
                     data->plist[ ici->tp1[t] ].z  );

         glVertex3d( data->plist[ ici->sp ].x,
                     data->plist[ ici->sp ].y,
                     data->plist[ ici->sp ].z  );
         glVertex3d( data->plist[ ici->tp2[t] ].x,
                     data->plist[ ici->tp2[t] ].y,
                     data->plist[ ici->tp2[t] ].z  );

         glVertex3d( data->plist[ ici->tp1[t] ].x,
                     data->plist[ ici->tp1[t] ].y,
                     data->plist[ ici->tp1[t] ].z  );
         glVertex3d( data->plist[ ici->tp2[t] ].x,
                     data->plist[ ici->tp2[t] ].y,
                     data->plist[ ici->tp2[t] ].z  );
      }

   }

   glEnd();
}
#endif


extern "C" void RHS_E_setup( const MainData *MD, void *data_,
                             string **omsg, int *nomsg )   // new omsg
{
   ModuleData *data = (ModuleData*)data_;
   const complex<double> cNULL(0.0,0.0);
   const complex<double> cONE(1.0,0.0);

   data->update_Iinc();

   delete [] data->IVrwg;
   data->IVrwg = new SEP_Z_VAL[ data->velist.size() ];


   triquad oI(3), iI(7);

   bool singu;
   int atri, qtri;
   double rv1[3], rv2[3], area;
   complex<double> A[3], phi, Gd[3][3], Gs, v[3], J[3], div, J0[3], div0;
   POINT ap, qp;

   complex<double> jomega( 0.0, MD->omega );

   for ( unsigned int e1=0; e1<data->velist.size(); ++e1 )
   {

   data->IVrwg[e1] = cNULL;

// triangle 1 p
   atri = data->velist[e1].trip;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].y
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].y
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].z
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].z
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].z;


// inc triangles
      for ( Iinc_LIST::iterator ici = data->Iinclist.begin();
            ici != data->Iinclist.end();
            ++ici )
      {

      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi = cNULL;

      for ( unsigned int itn=0; itn<ici->trino.size(); ++itn )
      {

      qtri = ici->trino[ itn ];

// edge 1
      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ ici->tp1[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( ap.y - data->plist[ ici->tp1[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( ap.z - data->plist[ ici->tp1[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( qp.x - data->plist[ ici->tp1[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( qp.y - data->plist[ ici->tp1[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( qp.z - data->plist[ ici->tp1[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ ici->tp1[itn] ].x ) * 0.5 * ici->angle[itn];
         rv2[1] = ( qp.y - data->plist[ ici->tp1[itn] ].y ) * 0.5 * ici->angle[itn];
         rv2[2] = ( qp.z - data->plist[ ici->tp1[itn] ].z ) * 0.5 * ici->angle[itn];

         div = cONE * ici->angle[itn];

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = ici->area[ itn ];
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

// edge 2
      if ( singu )
      {
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ ici->tp2[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( ap.y - data->plist[ ici->tp2[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( ap.z - data->plist[ ici->tp2[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( qp.x - data->plist[ ici->tp2[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( qp.y - data->plist[ ici->tp2[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( qp.z - data->plist[ ici->tp2[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
      }

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ ici->tp2[itn] ].x ) * 0.5 * ici->angle[itn];
         rv2[1] = ( qp.y - data->plist[ ici->tp2[itn] ].y ) * 0.5 * ici->angle[itn];
         rv2[2] = ( qp.z - data->plist[ ici->tp2[itn] ].z ) * 0.5 * ici->angle[itn];

         div = cONE * ici->angle[itn];

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = ici->area[ itn ];
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

      }  // end:itn

      rv1[0] = ( ap.x - data->plist[ data->velist[e1].spp ].x ) * 0.5;
      rv1[1] = ( ap.y - data->plist[ data->velist[e1].spp ].y ) * 0.5;
      rv1[2] = ( ap.z - data->plist[ data->velist[e1].spp ].z ) * 0.5;

      data->IVrwg[e1] += SEP_Z_VAL(
                         - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] ),
                         - 1.0 * phi / jomega
                                  ) * oI.weights[ip1] * ici->Jampl_fac;

      }  // end:ici

   }  // end:it 1 p

// triangle 1 m
   atri = data->velist[e1].trim;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].y
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].y
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].z
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].z
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].z;

// inc triangles
      for ( Iinc_LIST::iterator ici = data->Iinclist.begin();
            ici != data->Iinclist.end();
            ++ici )
      {

      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi = cNULL;

      for ( unsigned int itn=0; itn<ici->trino.size(); ++itn )
      {

      qtri = ici->trino[ itn ];

// edge 1
      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ ici->tp1[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( ap.y - data->plist[ ici->tp1[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( ap.z - data->plist[ ici->tp1[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( qp.x - data->plist[ ici->tp1[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( qp.y - data->plist[ ici->tp1[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( qp.z - data->plist[ ici->tp1[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ ici->tp1[itn] ].x ) * 0.5 * ici->angle[itn];
         rv2[1] = ( qp.y - data->plist[ ici->tp1[itn] ].y ) * 0.5 * ici->angle[itn];
         rv2[2] = ( qp.z - data->plist[ ici->tp1[itn] ].z ) * 0.5 * ici->angle[itn];

         div = cONE * ici->angle[itn];

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = ici->area[ itn ];
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

// edge 2
      if ( singu )
      {
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ ici->tp2[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( ap.y - data->plist[ ici->tp2[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( ap.z - data->plist[ ici->tp2[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
         else
         {
            data->triangle_center( qtri, qp );

            J0[0] = ( qp.x - data->plist[ ici->tp2[itn] ].x ) * 0.5 * ici->angle[itn];
            J0[1] = ( qp.y - data->plist[ ici->tp2[itn] ].y ) * 0.5 * ici->angle[itn];
            J0[2] = ( qp.z - data->plist[ ici->tp2[itn] ].z ) * 0.5 * ici->angle[itn];
            div0 = cONE * ici->angle[itn];
         }
      }

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ ici->tp2[itn] ].x ) * 0.5 * ici->angle[itn];
         rv2[1] = ( qp.y - data->plist[ ici->tp2[itn] ].y ) * 0.5 * ici->angle[itn];
         rv2[2] = ( qp.z - data->plist[ ici->tp2[itn] ].z ) * 0.5 * ici->angle[itn];

         div = cONE * ici->angle[itn];

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = ici->area[ itn ];
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

      }  // end:itn

      rv1[0] = ( data->plist[ data->velist[e1].spm ].x - ap.x ) * 0.5;
      rv1[1] = ( data->plist[ data->velist[e1].spm ].y - ap.y ) * 0.5;
      rv1[2] = ( data->plist[ data->velist[e1].spm ].z - ap.z ) * 0.5;

      data->IVrwg[e1] += SEP_Z_VAL(
                         - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] ),
                         - (-1.0) * phi / jomega
                                  ) * oI.weights[ip1] * ici->Jampl_fac;

      }  // end:ici

   }  // end:it 1 m

   }  // end:e1

   *nomsg = 0;
   *omsg = NULL;
}

//extern "C" void RHS_Einc( const MainData *MD, void *data, double x, double y, double z,
//                          complex<double> E[3] );

extern "C" complex<double> RHS_E_internal( const MainData *MD, void *data_, int ltno )   // see AIM
{
   ModuleData *data = (ModuleData*)data_;

   SEP_Z_VAL cum(0.0);
   LTbfkt *LTb = &data->LTbfktlist[ltno];

   for ( unsigned int i=0; i<LTb->count; ++i )
   {
      cum += data->IVrwg[ LTb->bfkt_in_LT[i] ] * LTb->norm[i];
   }

   if ( LTb->count > 1 ) return( cum.a_val );  // loop

   return( cum.a_val + cum.phi_val );
}

// new ampl, no appended to nobfkt
extern "C" void RHS_bfkt_inc( const MainData *MD, void *data_,
                              complex<double> **ampl, int *no )
{
   ModuleData *data = (ModuleData*)data_;

   *no = data->Iinclist.size();
   *ampl = new complex<double>[*no];
   for ( int i=0; i<*no; i++ )
   {
      (*ampl)[i] = data->Iinclist[i].Jampl_fac;
   }
}

extern "C" void RHS_E_cleanup( const MainData *MD, void *data_ )
{
   ModuleData *data = (ModuleData*)data_;
   delete [] data->IVrwg;
   data->IVrwg = NULL;
}





const int extract_para_no = 4;
const char *extract_para[ extract_para_no ] =
{
   "J VECT file",
   "no. of J probes",
   "no. of E probes",
   "no. of U probes"
};

const int P_para_no = 3;
const char *P_para[ P_para_no ] =
{
   "point x / m",
   "point y / m",
   "point z / m",
};

const int UP_para_no = 6;
const char *UP_para[ UP_para_no ] =
{
   "start point x / m",
   "start point y / m",
   "start point z / m",
   "end point x / m",  
   "end point y / m",  
   "end point z / m"
};


extern "C" void extract_get_parameter( const MainData *MD, void *data_,
                                       string **param, string **value,
                                       int *no )
{
   ModuleData *data = (ModuleData*)data_;
   const int buffsize = 50;
   char sb[buffsize];

   *no = extract_para_no + data->JPlist.size()*P_para_no
                         + data->EPlist.size()*P_para_no
                         + data->UPlist.size()*UP_para_no;
   *param = new string[*no];
   *value = new string[*no];

   int id = 0;
   int id2 = 0;

   (*param)[id] = extract_para[id2];
   (*value)[id] = data->J_vect_file_name;
   id++;


   id2 = 1;
   snprintf(sb,buffsize,"%u",(unsigned)data->JPlist.size());
   (*param)[id] = extract_para[id2];
   (*value)[id] = sb;
   id++;

   for ( unsigned int w=0; w<data->JPlist.size(); ++w )
   {
      id2 = 0;
      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->JPlist[w].p.x );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->JPlist[w].p.y );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->JPlist[w].p.z ); 
      (*value)[id] = sb;
      id2++;
      id++; 
   }

   id2 = 2;
   snprintf(sb,buffsize,"%u",(unsigned)data->EPlist.size());
   (*param)[id] = extract_para[id2];
   (*value)[id] = sb;
   id++;

   for ( unsigned int w=0; w<data->EPlist.size(); ++w )
   {
      id2 = 0;
      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->EPlist[w].p.x );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->EPlist[w].p.y );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, P_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->EPlist[w].p.z ); 
      (*value)[id] = sb;
      id2++;
      id++; 
   }


   id2 = 3;
   sprintf(sb,"%u",(unsigned)data->UPlist.size());
   (*param)[id] = extract_para[id2];
   (*value)[id] = sb;
   id++;

   for ( unsigned int w=0; w<data->UPlist.size(); ++w )
   {
      id2 = 0;
      snprintf( sb, buffsize, "%i:%s", w+1, UP_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->UPlist[w].p1.x );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, UP_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->UPlist[w].p1.y );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, UP_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->UPlist[w].p1.z ); 
      (*value)[id] = sb;
      id2++;
      id++;

      snprintf( sb, buffsize, "%i:%s", w+1, UP_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->UPlist[w].p2.x );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, UP_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->UPlist[w].p2.y );
      (*value)[id] = sb;
      id2++;
      id++; 

      snprintf( sb, buffsize, "%i:%s", w+1, UP_para[id2] );
      (*param)[id] = sb;
      snprintf( sb, buffsize, "%g", data->UPlist[w].p2.z ); 
      (*value)[id] = sb;
      id2++;
      id++; 
   }

}

extern "C" void extract_set_parameter( const MainData *MD, void *data_,
                                       const string *param, const string *value,
                                       int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j, nj, ne, nu;
   bool sec;
   const int buffsize = 50;
   char sb[buffsize];

   nj = data->JPlist.size();
   ne = data->EPlist.size();
   nu = data->UPlist.size();

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<extract_para_no; j++ ) if ( extract_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->J_vect_file_name = value[i]; break; }
         case 1: { nj = atoi( value[i].c_str() ); break; }
         case 2: { ne = atoi( value[i].c_str() ); break; }
         case 3: { nu = atoi( value[i].c_str() ); break; }
      }
   }

// Jprobe
   if ( nj != (int)data->JPlist.size() ) data->JPlist.resize(nj);

   for ( unsigned int g=0; g<data->JPlist.size(); ++g )
   {
      sec = true;
      for ( int i=0; i<no; i++ )
      {
         if ( extract_para[1] == param[i] )   // Jprobe
         {
            sec = false;
            continue;   
         }
         else if ( extract_para[2] == param[i] ) break;

         if ( sec ) continue;

         for ( j=0; j<P_para_no; j++ )
         {
            snprintf( sb, buffsize, "%i:%s", g+1, P_para[j] );

            if ( sb == param[i] )
            {
               switch(j)
               {
                  case 0: { data->JPlist[g].p.x = atof( value[i].c_str() ); break; }
                  case 1: { data->JPlist[g].p.y = atof( value[i].c_str() ); break; }
                  case 2: { data->JPlist[g].p.z = atof( value[i].c_str() ); break; }
               }
            }   

         }

      }

   }

// Eprobe
   if ( ne != (int)data->EPlist.size() ) data->EPlist.resize(ne);

   for ( unsigned int g=0; g<data->EPlist.size(); ++g )
   {
      sec = true;
      for ( int i=0; i<no; i++ )
      {
         if ( extract_para[2] == param[i] )   // Eprobe
         {
            sec = false;
            continue;   
         }
         else if ( extract_para[3] == param[i] ) break;

         if ( sec ) continue;

         for ( j=0; j<P_para_no; j++ )
         {
            snprintf( sb, buffsize, "%i:%s", g+1, P_para[j] );

            if ( sb == param[i] )
            {
               switch(j)
               {
                  case 0: { data->EPlist[g].p.x = atof( value[i].c_str() ); break; }
                  case 1: { data->EPlist[g].p.y = atof( value[i].c_str() ); break; }
                  case 2: { data->EPlist[g].p.z = atof( value[i].c_str() ); break; }
               }
            }   

         }

      }

   }

// Uprobe
   if ( nu != (int)data->UPlist.size() ) data->UPlist.resize(nu);

   for ( unsigned int g=0; g<data->UPlist.size(); ++g )
   {
      sec = true;
      for ( int i=0; i<no; i++ )
      {
         if ( extract_para[3] == param[i] )   // Uprobe
         {
            sec = false;
            continue;   
         }
         if ( sec ) continue;

         for ( j=0; j<UP_para_no; j++ )
         {
            snprintf( sb, buffsize, "%i:%s", g+1, UP_para[j] );

            if ( sb == param[i] )
            {
               switch(j)
               {
                  case 0: { data->UPlist[g].p1.x = atof( value[i].c_str() ); break; }
                  case 1: { data->UPlist[g].p1.y = atof( value[i].c_str() ); break; }
                  case 2: { data->UPlist[g].p1.z = atof( value[i].c_str() ); break; }
                  case 3: { data->UPlist[g].p2.x = atof( value[i].c_str() ); break; }
                  case 4: { data->UPlist[g].p2.y = atof( value[i].c_str() ); break; }
                  case 5: { data->UPlist[g].p2.z = atof( value[i].c_str() ); break; }
               }
            }   

         }

      }
   }

   data->setup_exvec();

}


#ifdef HAVE_LIBGL
extern "C" void extract_add_to_OGLlist( const MainData *MD, void *data_, GLuint list )
{
   const GLfloat clr_Jline[4]  = { 1.00, 0.30, 0.30, 1.0 };
   const GLfloat clr_Jcross[4] = { 0.30, 1.00, 0.30, 1.0 };
   const GLfloat clr_Ecross[4] = { 1.00, 0.00, 0.00, 1.0 };
   const GLfloat clr_1[4]   = { 1.00, 0.00, 1.00, 1.0 };
   const GLfloat clr_3[4]   = { 1.00, 0.00, 0.00, 1.0 };
   const GLfloat clr_4[4]   = { 0.00, 0.00, 1.00, 1.0 };
   ModuleData *data = (ModuleData*)data_;
   double R;
   POINT cp;

   R = 0.5 * HYPOT( MD->delta_x, MD->delta_y, MD->delta_z );

   for ( Jprobe_LIST::iterator jp = data->JPlist.begin();
         jp != data->JPlist.end();
         ++jp )
   {

// J line
      data->triangle_center( jp->tri, cp );
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_Jline );
      
      glBegin(GL_LINES);
      
      glVertex3d( jp->p.x, jp->p.y, jp->p.z );
      glVertex3d( cp.x, cp.y, cp.z );

// J triangle
      glVertex3d( data->plist[ data->tlist[jp->tri].p1 ].x,
                  data->plist[ data->tlist[jp->tri].p1 ].y,
                  data->plist[ data->tlist[jp->tri].p1 ].z  );
      glVertex3d( data->plist[ data->tlist[jp->tri].p2 ].x,
                  data->plist[ data->tlist[jp->tri].p2 ].y,
                  data->plist[ data->tlist[jp->tri].p2 ].z  );
      glVertex3d( data->plist[ data->tlist[jp->tri].p2 ].x,
                  data->plist[ data->tlist[jp->tri].p2 ].y,
                  data->plist[ data->tlist[jp->tri].p2 ].z  );
      glVertex3d( data->plist[ data->tlist[jp->tri].p3 ].x,
                  data->plist[ data->tlist[jp->tri].p3 ].y,
                  data->plist[ data->tlist[jp->tri].p3 ].z  );
      glVertex3d( data->plist[ data->tlist[jp->tri].p3 ].x,   
                  data->plist[ data->tlist[jp->tri].p3 ].y,   
                  data->plist[ data->tlist[jp->tri].p3 ].z  );
      glVertex3d( data->plist[ data->tlist[jp->tri].p1 ].x,
                  data->plist[ data->tlist[jp->tri].p1 ].y,
                  data->plist[ data->tlist[jp->tri].p1 ].z  );

      glEnd();

// J cross
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_Jcross );

      glBegin(GL_LINES);

      glVertex3d( jp->p.x - R, jp->p.y, jp->p.z );
      glVertex3d( jp->p.x + R, jp->p.y, jp->p.z );
      glVertex3d( jp->p.x, jp->p.y - R, jp->p.z );
      glVertex3d( jp->p.x, jp->p.y + R, jp->p.z );
      glVertex3d( jp->p.x, jp->p.y, jp->p.z - R );
      glVertex3d( jp->p.x, jp->p.y, jp->p.z + R );

      glEnd();
   }

// E cross
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_Ecross );

   glBegin(GL_LINES);

   for ( Eprobe_LIST::iterator ep = data->EPlist.begin();  
         ep != data->EPlist.end();
         ++ep )
   {
      glVertex3d( ep->p.x - R, ep->p.y, ep->p.z );
      glVertex3d( ep->p.x + R, ep->p.y, ep->p.z );
      glVertex3d( ep->p.x, ep->p.y - R, ep->p.z );
      glVertex3d( ep->p.x, ep->p.y + R, ep->p.z );
      glVertex3d( ep->p.x, ep->p.y, ep->p.z - R );
      glVertex3d( ep->p.x, ep->p.y, ep->p.z + R );
   }

   glEnd();


   for ( Uprobe_LIST::iterator up = data->UPlist.begin();
         up != data->UPlist.end();
         ++up )
   {
// point 1
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_3 );

      glBegin(GL_LINES);

      glVertex3d( up->p1.x - R, up->p1.y, up->p1.z );
      glVertex3d( up->p1.x + R, up->p1.y, up->p1.z );

      glVertex3d( up->p1.x, up->p1.y - R, up->p1.z );
      glVertex3d( up->p1.x, up->p1.y + R, up->p1.z );

      glVertex3d( up->p1.x, up->p1.y, up->p1.z - R );
      glVertex3d( up->p1.x, up->p1.y, up->p1.z + R );

      glEnd();

// point 2
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_4 );

      glBegin(GL_LINES);

      glVertex3d( up->p2.x - R, up->p2.y, up->p2.z );
      glVertex3d( up->p2.x + R, up->p2.y, up->p2.z );

      glVertex3d( up->p2.x, up->p2.y - R, up->p2.z );
      glVertex3d( up->p2.x, up->p2.y + R, up->p2.z );

      glVertex3d( up->p2.x, up->p2.y, up->p2.z - R );
      glVertex3d( up->p2.x, up->p2.y, up->p2.z + R );

      glEnd();

// line
      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_1 );

      glBegin(GL_LINES);

      glVertex3d( up->p1.x, up->p1.y, up->p1.z );
      glVertex3d( up->p2.x, up->p2.y, up->p2.z );

      glEnd();
   }

}
#endif


extern "C" void extract_run( const MainData *MD, void *data_, bool intern_inc,
                             complex<double> *JV, int no, const FIELD *f_,
                             string **omsg, int *nomsg )
{
   const complex<double> cNULL(0.0,0.0);
   ModuleData *data = (ModuleData*)data_;
   FIELDx *f = (FIELDx*)f_;
   string name;
   fstream jfile;
   complex<double> *JVrwg;

//
// back to RWG
   JVrwg = new complex<double>[no];
   for ( int i=0; i<no; ++i ) JVrwg[i] = cNULL;

   for ( LTbfkt_LIST::iterator lbi=data->LTbfktlist.begin();
         lbi != data->LTbfktlist.end();
         ++lbi )
   {
      for ( unsigned int i=0; i<lbi->count; ++i )
      {
#ifdef BOUNDARY_CHECK
         if ( lbi->bfkt_in_LT[i] < 0 || lbi->bfkt_in_LT[i] >= no ||
              lbi-data->LTbfktlist.begin() >= no                    )
         {
            cerr << "Error in LoopTreeModule extract_run(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }
#endif
         JVrwg[ lbi->bfkt_in_LT[i] ] += lbi->norm[i]
                                        * JV[ lbi-data->LTbfktlist.begin() ];
      }
   }

//
// Js VECT out
   if ( data->J_vect_file_name != "" )
   {
   name = MD->PrjPath + "/" + data->J_vect_file_name;

   double area, aj, maxlen, maxabs, nx, ny, nz;
   POINT cp;
   TRIANGLE *tri;

   complex<double> *Jx = new complex<double>[ data->tlist.size() ];
   complex<double> *Jy = new complex<double>[ data->tlist.size() ];
   complex<double> *Jz = new complex<double>[ data->tlist.size() ];

   maxlen = 0.0;
   maxabs = 0.0;

   for ( unsigned int t=0; t<data->tlist.size(); ++t )
   {
      data->triangle_center( t, cp );
      area = data->triangle_area( t );
      tri = &data->tlist[t];

      Jx[t] = cNULL;
      Jy[t] = cNULL;
      Jz[t] = cNULL;

// edge 1
      Jx[t] += JVrwg[ tri->ve1 ] * (double)tri->sign.sp1
               * ( cp.x - data->plist[ tri->p1 ].x );
      Jy[t] += JVrwg[ tri->ve1 ] * (double)tri->sign.sp1
               * ( cp.y - data->plist[ tri->p1 ].y );
      Jz[t] += JVrwg[ tri->ve1 ] * (double)tri->sign.sp1
               * ( cp.z - data->plist[ tri->p1 ].z );

// edge 2
      Jx[t] += JVrwg[ tri->ve2 ] * (double)tri->sign.sp2
               * ( cp.x - data->plist[ tri->p2 ].x );
      Jy[t] += JVrwg[ tri->ve2 ] * (double)tri->sign.sp2
               * ( cp.y - data->plist[ tri->p2 ].y );
      Jz[t] += JVrwg[ tri->ve2 ] * (double)tri->sign.sp2
               * ( cp.z - data->plist[ tri->p2 ].z );

// edge 3
      Jx[t] += JVrwg[ tri->ve3 ] * (double)tri->sign.sp3
               * ( cp.x - data->plist[ tri->p3 ].x );
      Jy[t] += JVrwg[ tri->ve3 ] * (double)tri->sign.sp3
               * ( cp.y - data->plist[ tri->p3 ].y );
      Jz[t] += JVrwg[ tri->ve3 ] * (double)tri->sign.sp3
               * ( cp.z - data->plist[ tri->p3 ].z );

      Jx[t] /= 2.0*area;
      Jy[t] /= 2.0*area;
      Jz[t] /= 2.0*area;

      aj = abs(sqrt( Jx[t]*conj(Jx[t]) + Jy[t]*conj(Jy[t]) + Jz[t]*conj(Jz[t]) ));
      if ( maxabs < aj ) maxabs = aj;

      maxlen += area;
   }
   maxlen /= data->tlist.size();
   maxlen = 1.0 * sqrt( 2.0*maxlen );

   if ( intern_inc ) 
   {
      for ( unsigned int j=0; j<data->Iinclist.size(); ++j )
      {
         Iinc *inc = &(data->Iinclist[j]);

         for ( unsigned int t=0; t<inc->trino.size(); ++t )
         {
            data->triangle_center( inc->trino[t], cp );

            Jx[ inc->trino[t] ]
               += data->Iinclist[j].Jampl_fac
                  * 0.5 * inc->angle[t]
                  * ( cp.x - data->plist[ inc->tp1[t] ].x )
                  / inc->area[t];

            Jy[ inc->trino[t] ]
               += data->Iinclist[j].Jampl_fac
                  * 0.5 * inc->angle[t]
                  * ( cp.y - data->plist[ inc->tp1[t] ].y )
                  / inc->area[t];

            Jz[ inc->trino[t] ]
               += data->Iinclist[j].Jampl_fac
                  * 0.5 * inc->angle[t]
                  * ( cp.z - data->plist[ inc->tp1[t] ].z )
                  / inc->area[t];


            Jx[ inc->trino[t] ]
               += data->Iinclist[j].Jampl_fac
                  * 0.5 * inc->angle[t]
                  * ( cp.x - data->plist[ inc->tp2[t] ].x )
                  / inc->area[t];

            Jy[ inc->trino[t] ]
               += data->Iinclist[j].Jampl_fac
                  * 0.5 * inc->angle[t]
                  * ( cp.y - data->plist[ inc->tp2[t] ].y )
                  / inc->area[t];

            Jz[ inc->trino[t] ]
               += data->Iinclist[j].Jampl_fac
                  * 0.5 * inc->angle[t]
                  * ( cp.z - data->plist[ inc->tp2[t] ].z )
                  / inc->area[t];

         }

      }
   }


   double R;
   POLY p;
   POLYLIST pl;
   pl.clear();

   R = maxlen / maxabs;

   for ( unsigned int t=0; t<data->tlist.size(); ++t )  
   {
      data->triangle_center( t, cp );

      p = POLY( 2, 255, 0, 0 );
      p.x[0] = cp.x;
      p.y[0] = cp.y;
      p.z[0] = cp.z;

      p.x[1] = p.x[0] + Jx[t].real()*R;
      p.y[1] = p.y[0] + Jy[t].real()*R;
      p.z[1] = p.z[0] + Jz[t].real()*R;

      pl.push_back(p);


      p = POLY( 2, 0, 0, 255 );
      p.x[0] = cp.x;
      p.y[0] = cp.y;
      p.z[0] = cp.z;

      p.x[1] = p.x[0] + Jx[t].imag()*R;
      p.y[1] = p.y[0] + Jy[t].imag()*R;
      p.z[1] = p.z[0] + Jz[t].imag()*R;

      pl.push_back(p);


      data->triangle_normal( t, nx, ny, nz );

      p = POLY( 2, 255, 0, 255 );
      p.x[0] = cp.x - 0.05*maxlen*nx;
      p.y[0] = cp.y - 0.05*maxlen*ny;
      p.z[0] = cp.z - 0.05*maxlen*nz;

      p.x[1] = cp.x + 0.05*maxlen*nx;
      p.y[1] = cp.y + 0.05*maxlen*ny;
      p.z[1] = cp.z + 0.05*maxlen*nz;

      pl.push_back(p);

   }

   name = MD->PrjPath + "/" + data->J_vect_file_name;
   vect_out( name, &pl );

   delete [] Jz;
   delete [] Jy;
   delete [] Jx;
   }


//
// Js probe out (without inc)
   vector< complex<double> > JX, JY, JZ;

   JX.clear();
   JY.clear();
   JZ.clear();

   int trino;
   POINT pkt;
   double area, nx, ny, nz;
   complex<double> Jx, Jy, Jz;
   TRIANGLE *tri;

   for ( Jprobe_LIST::iterator jp = data->JPlist.begin();  
         jp != data->JPlist.end();
         ++jp )
   {
      Jx = Jy = Jz = cNULL;

      trino = jp->tri;
      pkt = jp->p;

      area = data->triangle_normal( trino, nx, ny, nz );

      tri = &( data->tlist[trino] );
// edge 1
      if ( tri->sign.sp1 != 0 )
      {
         Jx += JVrwg[ tri->ve1 ] * (double)tri->sign.sp1
               * ( pkt.x - data->plist[ tri->p1 ].x ) * 0.5 / area;
         Jy += JVrwg[ tri->ve1 ] * (double)tri->sign.sp1
               * ( pkt.y - data->plist[ tri->p1 ].y ) * 0.5 / area;
         Jz += JVrwg[ tri->ve1 ] * (double)tri->sign.sp1
               * ( pkt.z - data->plist[ tri->p1 ].z ) * 0.5 / area;
      }
// edge 2
      if ( tri->sign.sp2 != 0 )
      {
         Jx += JVrwg[ tri->ve2 ] * (double)tri->sign.sp2
               * ( pkt.x - data->plist[ tri->p2 ].x ) * 0.5 / area;
         Jy += JVrwg[ tri->ve2 ] * (double)tri->sign.sp2
               * ( pkt.y - data->plist[ tri->p2 ].y ) * 0.5 / area;
         Jz += JVrwg[ tri->ve2 ] * (double)tri->sign.sp2
               * ( pkt.z - data->plist[ tri->p2 ].z ) * 0.5 / area;
      }
// edge 3
      if ( tri->sign.sp3 != 0 )
      {
         Jx += JVrwg[ tri->ve3 ] * (double)tri->sign.sp3
               * ( pkt.x - data->plist[ tri->p3 ].x ) * 0.5 / area;
         Jy += JVrwg[ tri->ve3 ] * (double)tri->sign.sp3
               * ( pkt.y - data->plist[ tri->p3 ].y ) * 0.5 / area;
         Jz += JVrwg[ tri->ve3 ] * (double)tri->sign.sp3
               * ( pkt.z - data->plist[ tri->p3 ].z ) * 0.5 / area;
      }

      JX.push_back(Jx);
      JY.push_back(Jy);
      JZ.push_back(Jz);
   }

//
// E probe out
   complex<double> E[3], E_t[3];
   vector< complex<double> > EincX, EincY, EincZ, EscattX, EscattY, EscattZ;

   bool outside;

   EincX.clear();
   EincY.clear();
   EincZ.clear();
   EscattX.clear();
   EscattY.clear();
   EscattZ.clear();

   for ( Eprobe_LIST::iterator ep = data->EPlist.begin();
         ep != data->EPlist.end();
         ++ep )
   {
// location
      if ( MD->cpos_x+0.5*MD->size_x+0.001*MD->delta_x < ep->p.x ||
           MD->cpos_x-0.5*MD->size_x-0.001*MD->delta_x > ep->p.x ||
           MD->cpos_y+0.5*MD->size_y+0.001*MD->delta_y < ep->p.y ||
           MD->cpos_y-0.5*MD->size_y-0.001*MD->delta_y > ep->p.y ||
           MD->cpos_z+0.5*MD->size_z+0.001*MD->delta_z < ep->p.z ||
           MD->cpos_z-0.5*MD->size_z-0.001*MD->delta_z > ep->p.z    )
      {
         outside = true;
      }
      else
      {
         outside = false;
      }

//
// scatt
      E[0] = E[1] = E[2] = cNULL;

// intern
      for ( unsigned int j=0; j<data->velist.size(); j++ )
      {
         data->bfkt_E( E_t, j, ep->p );
         E[0] += JVrwg[j] * E_t[0];
         E[1] += JVrwg[j] * E_t[1];
         E[2] += JVrwg[j] * E_t[2];
      }

// extern
      if ( outside )
      {
         f->E_Jtotal_sum( *data->grf, ep->p, E_t );
         E[0] += E_t[0];
         E[1] += E_t[1];
         E[2] += E_t[2];
      }
      else
      {
         f->E_scatt( ep->p, E_t );
         E[0] += E_t[0];
         E[1] += E_t[1];
         E[2] += E_t[2];
      }

      EscattX.push_back( E[0] );
      EscattY.push_back( E[1] );
      EscattZ.push_back( E[2] );

//
// inc
      E[0] = E[1] = E[2] = cNULL;

// intern
      if ( intern_inc )
      {
         for ( unsigned int j=0; j<data->Iinclist.size(); j++ )
         {
            data->inc_E( E_t, j, ep->p );
            E[0] += data->Iinclist[j].Jampl_fac * E_t[0];    
            E[1] += data->Iinclist[j].Jampl_fac * E_t[1];  
            E[2] += data->Iinclist[j].Jampl_fac * E_t[2];  
         }
      }

// extern
      if ( outside )
      {
// NOT AVAILABLE!
      }
      else
      {
         f->E_inc( ep->p, E_t );
         E[0] += E_t[0];
         E[1] += E_t[1];
         E[2] += E_t[2];
      }
      

      EincX.push_back( E[0] );
      EincY.push_back( E[1] );
      EincZ.push_back( E[2] );
   }


//
// U probe out (internal only!)
   vector< complex<double> > U;

   U.clear();

   const int noil = 16;

   double *lw = new double[noil];
   double *lx = new double[noil];
   GAUSS( const_cast<int*>(&noil), lw, lx );
   for ( int i=0; i<noil; i++ )
   {
      lx[i] = 0.5 * ( lx[i] + 1.0 );
      lw[i] *= 0.5;
   }

   int noint;
   double l12, lv[3], sgr;
   complex<double> A[3], A_t[3], PHI1, PHI2, PHI_t, U_t;
   POINT AP, P1, P2;

   complex<double> jomega( 0.0, MD->omega );

   sgr = 0.1 * MD->wavelength;

   for ( Uprobe_LIST::iterator up = data->UPlist.begin();
         up != data->UPlist.end();
         ++up )
   {
      U_t = complex<double>(0.0,0.0);

      lv[0] = up->p2.x - up->p1.x;
      lv[1] = up->p2.y - up->p1.y;
      lv[2] = up->p2.z - up->p1.z;

      l12 = HYPOT( lv[0], lv[1], lv[2] );

      noint = (int)ceil( 8.0*l12/MD->wavelength );

      for ( int ii=0; ii<noint; ii++ )
      {
      P1.x = up->p1.x + ii * lv[0] / noint;
      P1.y = up->p1.y + ii * lv[1] / noint;
      P1.z = up->p1.z + ii * lv[2] / noint;

      P2.x = up->p1.x + (ii+1) * lv[0] / noint;
      P2.y = up->p1.y + (ii+1) * lv[1] / noint;
      P2.z = up->p1.z + (ii+1) * lv[2] / noint;

      for ( int ig=0; ig<noil; ig++ )
      {
         AP.x = P1.x + lx[ig] * ( P2.x - P1.x );
         AP.y = P1.y + lx[ig] * ( P2.y - P1.y );
         AP.z = P1.z + lx[ig] * ( P2.z - P1.z );

// intern
         A[0] = A[1] = A[2] = cNULL;
         for ( unsigned int j=0; j<data->velist.size(); ++j )
         {
            data->bfkt_pot( A_t, PHI_t, sgr, j, AP );
            A[0] += JVrwg[j] * A_t[0];
            A[1] += JVrwg[j] * A_t[1];
            A[2] += JVrwg[j] * A_t[2];
         }
         if ( intern_inc )
         {
            for ( unsigned int j=0; j<data->Iinclist.size(); ++j )
            {
               data->bfkt_pot( A_t, PHI_t, sgr, data->velist.size()+j, AP );
               A[0] += data->Iinclist[j].Jampl_fac * A_t[0];
               A[1] += data->Iinclist[j].Jampl_fac * A_t[1];
               A[2] += data->Iinclist[j].Jampl_fac * A_t[2];
            }
         }

// no extern!
         E[0] = E[1] = E[2] = cNULL;


         E[0] -= jomega * A[0];
         E[1] -= jomega * A[1];
         E[2] -= jomega * A[2];

         U_t += lw[ig] * ( E[0]*lv[0] + E[1]*lv[1] + E[2]*lv[2] ) / (double)noint;
      }
      }


      PHI1 = PHI2 = cNULL;
      AP = up->p2;
      for ( LTbfkt_LIST::iterator it = data->LTbfktlist.begin();
            it != data->LTbfktlist.end();
            ++it )
      {
         if ( it->count != 1 ) continue;  // tree only

         data->bfkt_pot( A_t, PHI_t, sgr, it->bfkt_in_LT[0], AP );
         PHI2 += JV[ it - data->LTbfktlist.begin() ] * PHI_t * it->norm[0];
      }
      if ( intern_inc )
      {
         for ( unsigned int j=0; j<data->Iinclist.size(); ++j )
         {
            data->bfkt_pot( A_t, PHI_t, sgr, data->velist.size()+j, AP );
            PHI2 += data->Iinclist[j].Jampl_fac * PHI_t;
         }
      }

      AP = up->p1;
      for ( LTbfkt_LIST::iterator it = data->LTbfktlist.begin();
            it != data->LTbfktlist.end();
            ++it )
      {
         if ( it->count != 1 ) continue;  // tree only

         data->bfkt_pot( A_t, PHI_t, sgr, it->bfkt_in_LT[0], AP );
         PHI1 += JV[ it - data->LTbfktlist.begin() ] * PHI_t * it->norm[0];
      }
      if ( intern_inc )
      {
         for ( unsigned int j=0; j<data->Iinclist.size(); ++j )
         {
            data->bfkt_pot( A_t, PHI_t, sgr, data->velist.size()+j, AP );
            PHI1 += data->Iinclist[j].Jampl_fac * PHI_t;
         }
      }

      U_t -= - ( PHI2 - PHI1 ) / jomega;


      U.push_back(U_t);
   }

   delete [] lx;
   delete [] lw;

   delete [] JVrwg;

//
// generate messages

   const int buffsize = 64;
   char buff[buffsize];

   int nom = 0;

   *nomsg = 4*JX.size() + 10*EincX.size()+ U.size();
   *omsg = new string[*nomsg];

   for ( unsigned int j=0; j<JX.size(); j++ )
   {
      snprintf( buff, buffsize, "Js probe position no. %i: ( %g; %g; %g ) m",
                j+1, data->JPlist[j].p.x,
                        data->JPlist[j].p.y,
                           data->JPlist[j].p.z );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(J_x) = %g A/m; arg(J_x) = %g°",
                abs(JX[j]), arg(JX[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(J_y) = %g A/m; arg(J_y) = %g°",
                abs(JY[j]), arg(JY[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(J_z) = %g A/m; arg(J_z) = %g°",
                abs(JZ[j]), arg(JZ[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;
   }

   for ( unsigned int j=0; j<EincX.size(); j++ )
   {
      snprintf( buff, buffsize, "E probe position no. %i: ( %g; %g; %g ) m",
                j+1, data->EPlist[j].p.x,
                        data->EPlist[j].p.y,
                           data->EPlist[j].p.z );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Einc_x) = %g V/m; arg(Einc_x) = %g°",
                abs(EincX[j]), arg(EincX[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Einc_y) = %g V/m; arg(Einc_y) = %g°",
                abs(EincY[j]), arg(EincY[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Einc_z) = %g V/m; arg(Einc_z) = %g°",
                abs(EincZ[j]), arg(EincZ[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Escatt_x) = %g V/m; arg(Escatt_x) = %g°",
                abs(EscattX[j]), arg(EscattX[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Escatt_y) = %g V/m; arg(Escatt_y) = %g°",
                abs(EscattY[j]), arg(EscattY[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Escatt_z) = %g V/m; arg(Escatt_z) = %g°",
                abs(EscattZ[j]), arg(EscattZ[j])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      E_t[0] = EincX[j] + EscattX[j];
      E_t[1] = EincY[j] + EscattY[j];
      E_t[2] = EincZ[j] + EscattZ[j];

      snprintf( buff, buffsize, "  abs(Etotal_x) = %g V/m; arg(Etotal_x) = %g°",
                abs(E_t[0]), arg(E_t[0])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Etotal_y) = %g V/m; arg(Etotal_y) = %g°",
                abs(E_t[1]), arg(E_t[1])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

      snprintf( buff, buffsize, "  abs(Etotal_z) = %g V/m; arg(Etotal_z) = %g°",
                abs(E_t[2]), arg(E_t[2])*180.0/M_PI );
      (*omsg)[nom] = buff;
      nom++;

   }

   for ( unsigned int g=0; g<U.size(); g++ )
   {
      snprintf( buff, buffsize, "voltage at probe no. %i: ( %g; %g ) V", g+1, U[g].real(), U[g].imag() );
      (*omsg)[nom] = buff;
      nom++;
   }

}
