!-----------------------------------------------------------------------
!--- Gaussquadratur fr Integration von x = -1 nach x = +1
!--- Version: (19.6.2003) ----------------------------------------------
!-----------------------------------------------------------------------
!--- nach ,,Numerical Recipes''
!--- n : Grad der Quadratur
!--- w(1...n) : Gewichte
!--- x(1...n) : Abszissenwerte
      subroutine gauss(n,w,x)

      implicit none

      real*8 pi, error
      parameter( pi = 3.14159265358979323846264338327d0 )
      parameter( error = 1d-15 )

      integer i,j,m
      integer n
      real*8 w(1:n),x(1:n)
      real*8 p1,p2,p3,pp,z,z1

      m = (n+1)/2
      z1 = 0d0
      index_i: do i=1, m
       z = cos( pi*( i - 0.25d0 ) / ( n + 0.5d0 ) )
       iteration: do
        p1 = 1d0
        p2 = 0d0
        index_j: do j=1, n
         p3 = p2
         p2 = p1
         p1 = ( ( 2d0*j - 1d0 )*z*p2 - ( j - 1d0 )*p3 )/j
        end do index_j
        pp = n*(z*p1-p2)/(z*z-1d0)
        z1 = z
        z  = z1 - p1/pp
        if ( abs(z-z1) < error ) exit iteration
       end do iteration
       x(i)     =  -z
       x(n+1-i) =   z
       w(i)     = 2d0/( ( 1d0 - z*z )*pp*pp )
       w(n+1-i) = w(i)
      end do index_i

      end subroutine gauss
