//
// module data
// (c) 18.4.2007 Ingo D. Rullhusen
//

#ifndef MODULE_DATA_H
#define MODULE_DATA_H

#include <GL/gl.h>

#include <string>

#include "point.h"
#include "green.h"
#include "sparse_matrix.h"
#include "sep_val.h"

using namespace std;


struct TRIANGLE
{
   int p1, p2, p3;
   int ve1, ve2, ve3;
   struct
   {
      signed short sp1:2;  // -1; 0; +1
      signed short sp2:2;  // -1; 0; +1
      signed short sp3:2;  // -1; 0; +1
   } sign;
};

typedef vector<TRIANGLE> TRIANGLE_LIST;


struct vEDGE
{
   int p1, p2;
   int spp, spm;
   int trip, trim;
};

typedef vector<vEDGE> vEDGE_LIST;


class LTbfkt
{
   public:
      unsigned int count;
      vector<double> norm;
      vector<int> bfkt_in_LT;
};

typedef vector<LTbfkt> LTbfkt_LIST;


class Iinc
{
   public:
      double x, y, z;
      complex<double> Iampl;

      int sp, count;
      vector<int> tp1, tp2, trino;
      vector<double> angle, area;
      complex<double> Jampl_fac;

      Iinc(){ count = 0; }
};

typedef vector<Iinc> Iinc_LIST;


class Jprobe
{
   public:
      int tri;
      POINT p;
};

typedef vector<Jprobe> Jprobe_LIST;

class Eprobe
{
   public:
      POINT p;
};

typedef vector<Eprobe> Eprobe_LIST;

class Uprobe
{
   public:
      POINT p1, p2;
};

typedef vector<Uprobe> Uprobe_LIST;


class ModuleData
{
   public:

// matrix
      string off_file_name;
      double mesh_unit;
      double cp_x, cp_y, cp_z;
      double theta, phi, psi;

      GREEN *grf;

      POINT_LIST     plist;
      TRIANGLE_LIST  tlist;
      vEDGE_LIST     velist;

      LTbfkt_LIST LTbfktlist;

// RHS
      Iinc_LIST Iinclist;

// extract
      string J_vect_file_name;

      Jprobe_LIST JPlist;
      Eprobe_LIST EPlist;
      Uprobe_LIST UPlist;


      ModuleData();
      ModuleData( const ModuleData *d );
      ~ModuleData();

      void read_off_file( string &msg );
#ifdef HAVE_LIBGL
      void ogl_triangles( GLuint list );
#endif

      void triangle_center( int tno, POINT &p ) const;
      bool triangle_neigh( int tno1, int tno2 ) const;
      double triangle_area( int tno ) const;
      double triangle_normal( int tno, double &nx, double &ny, double &nz ) const;
      int near_triangle( POINT &p ) const;

      void umkugel( int ltno, double &R, double CP[3] ) const;

      void bfkt_E( complex<double> E[3], int j, POINT p ) const;
// j in RWG basis
      void inc_E( complex<double> E[3], int j, POINT p ) const;
// j >= velist.size()
      void bfkt_pot( complex<double> A[3], complex<double> &mjoPHI,
                     double sgr, int j, POINT AP );
// j in RWG basis or >= velist.size()


      void update_Iinc();

      void setup_exvec();

// temporary
      SparseMatrix<SEP_Z_VAL> *Zrwg;
      SEP_Z_VAL *IVrwg;

};


#endif
