//
// Function for analytical singularity integration
// (c) 2.10.2003 Ingo D. Rullhusen
//

#include <math.h>
#include <float.h>


static double SQU( double x ){ return(x*x); }


double Psing( int n, double *bx, double *by,
              double rx, double ry, double rz )
{
   int i;
   double sg, nv, sig, p0i, ui[2], li[2], rp[3], rm[3], r0[3],
          rpi, rmi, r0iq, lpi, lmi, logt, d;

   sg = 0.0;

   for ( i=0; i<n-1; i++ )
   {

      li[0] = bx[i+1] - bx[i];
      li[1] = by[i+1] - by[i];
      nv = hypot( li[0], li[1] );
      li[0] /= nv;
      li[1] /= nv;

      ui[0] =  li[1];
      ui[1] = -li[0];

      rp[0] = bx[i+1] - rx;
      rp[1] = by[i+1] - ry;
      rp[2] =         - rz;

      rpi = sqrt( SQU(rp[0]) + SQU(rp[1]) + SQU(rp[2]) );
      lpi = rp[0]*li[0] + rp[1]*li[1];

      rm[0] = bx[i] - rx;
      rm[1] = by[i] - ry;
      rm[2] =       - rz;

      rmi = sqrt( SQU(rm[0]) + SQU(rm[1]) + SQU(rm[2]) );
      lmi = rm[0]*li[0] + rm[1]*li[1];

      nv = rm[0]*li[0] + rm[1]*li[1];
      r0[0] = rm[0] - nv * li[0];
      r0[1] = rm[1] - nv * li[1];
      r0[2] = rm[2];

      r0iq = SQU(r0[0]) + SQU(r0[1]) + SQU(r0[2]);

      p0i = ui[0]*rm[0] + ui[1]*rm[1];
      sig = ( p0i < 0.0 ? -1.0 : +1.0 );
      p0i = fabs( p0i );

      d = fabs(rz);

      logt = 0.0;
      if ( fabs(rpi+lpi) > DBL_EPSILON ) logt += p0i * log( rpi + lpi );
      if ( fabs(rmi+lmi) > DBL_EPSILON ) logt -= p0i * log( rmi + lmi );

      sg += sig * ( logt - d * ( atan2( p0i*lpi, r0iq+d*rpi )
                                 - atan2( p0i*lmi, r0iq+d*rmi ) ) );

   }

   return(sg);
}

double Hsing( int n, double *bx, double *by,
              double rx, double ry, double rz )
{
   int i;
   double sg, nv, sig, p0i, ui[2], li[2], rp[3], rm[3], r0[3],
          rpi, rmi, r0iq, lpi, lmi, ppi, pmi, logp, logm;

   sg = 0.0;

   for ( i=0; i<n-1; i++ )
   {

      li[0] = bx[i+1] - bx[i];
      li[1] = by[i+1] - by[i];
      nv = hypot( li[0], li[1] );
      li[0] /= nv;
      li[1] /= nv;

      ui[0] =  li[1];
      ui[1] = -li[0];

      rp[0] = bx[i+1] - rx;
      rp[1] = by[i+1] - ry;
      rp[2] =         - rz;

      rpi = sqrt( SQU(rp[0]) + SQU(rp[1]) + SQU(rp[2]) );
      lpi = rp[0]*li[0] + rp[1]*li[1];

      rm[0] = bx[i] - rx;
      rm[1] = by[i] - ry;
      rm[2] =       - rz;

      rmi = sqrt( SQU(rm[0]) + SQU(rm[1]) + SQU(rm[2]) );
      lmi = rm[0]*li[0] + rm[1]*li[1];

      nv = rm[0]*li[0] + rm[1]*li[1];
      r0[0] = rm[0] - nv * li[0];
      r0[1] = rm[1] - nv * li[1];
      r0[2] = rm[2];

      r0iq = SQU(r0[0]) + SQU(r0[1]) + SQU(r0[2]);

      p0i = ui[0]*rm[0] + ui[1]*rm[1];
      sig = ( p0i < 0.0 ? -1.0 : +1.0 );
      p0i = fabs( p0i );

      ppi = hypot( rp[0], rp[1] );
      pmi = hypot( rm[0], rm[1] );

      if ( fabs(ppi) > DBL_EPSILON )
         logp = lpi*log(ppi);
      else
         logp = 0.0;

      if ( fabs(pmi) > DBL_EPSILON )
         logm = lmi*log(pmi);
      else
         logm = 0.0;

      sg += sig*p0i * ( logp - logm
                        + p0i*( atan2( lpi, p0i ) - atan2( lmi, p0i ) )
                        - 1.5 * ( lpi - lmi )
                      );

   }

   return(0.5*sg);
}
